#include "array.h"
#include "../../common/aliases/aliases.h"
#include "../../common/misc/misc_utils.h"
#include "../../core/mem/allocator/mem_allocator.h"
#include <stdlib.h>

i32 *wapp_i32_array_get(const I32Array *array, u64 index) {
  if (!array || index >= array->count) {
    return NULL;
  }

  return &(array->items[index]);
}

void wapp_i32_array_set(I32Array *array, u64 index, i32 item) {
  if (!array || index >= array->count) {
    return;
  }

  array->items[index] = item;
}

void wapp_i32_array_append_capped(I32Array *array, i32 item) {
  if (!array || array->count >= array->capacity) {
    return;
  }

  array->items[(array->count)++] = item;
}

void wapp_i32_array_extend_capped(I32Array *array, const I32Array *other) {
  if (!array || !other) {
    return;
  }

  u64 remaining_capacity = array->capacity - array->count;
  if (other->count >= remaining_capacity) {
    return;
  }

  i32 *item;
  u64 items_to_add = other->count;
  for (u64 i = 0; i < items_to_add; ++i) {
    item = wapp_i32_array_get(other, i);
    if (!item) {
      continue;
    }

    wapp_i32_array_append_capped(array, *item);
  }
}

void wapp_i32_array_clear(I32Array *array) {
  if (!array) {
    return;
  }

  array->count = 0;
}

i32 wapp_i32_array_pop(I32Array *array) {
  if (!array || array->count == 0) {
    return (i32){0};
  }

  return array->items[--(array->count)];
}

void wapp_i32_array_copy_capped(const I32Array *src, I32Array *dst) {
  if (!src || !dst) {
    return;
  }

  wapp_i32_array_clear(dst);

  i32 *item;
  u64 to_copy = src->count < dst->capacity ? src->count : dst->capacity;
  for (u64 i = 0; i < to_copy; ++i) {
    item = wapp_i32_array_get(src, i);
    if (!item) {
      continue;
    }

    wapp_i32_array_append_capped(dst, *item);
  }
}

I32Array *wapp_i32_array_alloc_capacity(const Allocator *allocator, u64 capacity) {
  u64 allocation_size = sizeof(I32Array) + sizeof(i32) * capacity;
  I32Array *array     = NULL;

  if (!allocator) {
    goto RETURN_I32_ARRAY_ALLOC;
  }

  array = wapp_mem_allocator_alloc(allocator, allocation_size);
  if (!array) {
    goto RETURN_I32_ARRAY_ALLOC;
  }

  array->items    = (i32 *)((u8 *)array + sizeof(I32Array));
  array->count    = 0;
  array->capacity = capacity;

RETURN_I32_ARRAY_ALLOC:
  return array;
}

I32Array *wapp_i32_array_append_alloc(const Allocator *allocator, I32Array *array, i32 item) {
  I32Array *output = array;

  if (!allocator || !array) {
    goto RETURN_I32_ARRAY_APPEND_ALLOC;
  }

  if (array->count >= array->capacity) {
    u64 new_capacity = wapp_misc_utils_u64_round_up_pow2(array->capacity * 2);
    output           = wapp_i32_array_alloc_capacity(allocator, new_capacity);
    if (!output) {
      output = array;
      goto RETURN_I32_ARRAY_APPEND_ALLOC;
    }
    wapp_i32_array_copy_capped(array, output);
  }

  wapp_i32_array_append_capped(output, item);

RETURN_I32_ARRAY_APPEND_ALLOC:
  return output;
}

I32Array *wapp_i32_array_extend_alloc(const Allocator *allocator, I32Array *array, const I32Array *other) {
  I32Array *output = array;

  if (!allocator || !array || !other) {
    goto RETURN_I32_ARRAY_EXTEND_ALLOC;
  }

  u64 remaining_capacity = array->capacity - array->count;
  if (other->count >= remaining_capacity) {
    u64 new_capacity = wapp_misc_utils_u64_round_up_pow2(array->capacity * 2);
    output           = wapp_i32_array_alloc_capacity(allocator, new_capacity);
    if (!output) {
      output = array;
      goto RETURN_I32_ARRAY_EXTEND_ALLOC;
    }
    wapp_i32_array_copy_capped(array, output);
  }

  wapp_i32_array_extend_capped(output, other);

RETURN_I32_ARRAY_EXTEND_ALLOC:
  return output;
}

I32Array *wapp_i32_array_copy_alloc(const Allocator *allocator, const I32Array *src, I32Array *dst) {
  I32Array *output = dst;

  if (!allocator || !src || !dst) {
    goto RETURN_I32_ARRAY_COPY_ALLOC;
  }

  if (src->count >= dst->capacity) {
    u64 new_capacity = wapp_misc_utils_u64_round_up_pow2(dst->capacity * 2);
    output           = wapp_i32_array_alloc_capacity(allocator, new_capacity);
    if (!output) {
      output = dst;
      goto RETURN_I32_ARRAY_COPY_ALLOC;
    }
  }

  wapp_i32_array_clear(output);
  wapp_i32_array_copy_capped(src, output);

RETURN_I32_ARRAY_COPY_ALLOC:
  return output;
}
